#ifndef PATHFINDER_H
#define PATHFINDER_H

#include <vector>
#include <utility>
#include <string>
#include <cmath>
#include <algorithm>

#define DIAG_DIST 1.41421
#define EPS 0.0000001

class Position {
public:
	int x;
	int y;

	Position(int x, int y) :
			x(x), y(y){
	}

	static Position fromIndex(int index, int width){
		return Position(index % width, index / width);
	}

	inline int toIndex(int width){
		return y * width + x;
	}

	bool operator<(const Position &pos) const{
		return x < pos.x || (x == pos.x && y < pos.y);
	}

	bool operator==(const Position &pos) const{
		return x == pos.x && y == pos.y;
	}
};

class Pathfinder {
public:
	virtual void preprocess(std::vector<int> &map, int width, int height) = 0;
	virtual void findPath(Position start, Position goal, std::vector<Position> &path,
			std::vector<int> &map, int width, int height) = 0;
	virtual void clear() = 0;
	virtual std::string getName() = 0;
	virtual size_t getSize() = 0;

	std::vector<Position> getRoots(){
		return roots;
	}

	static std::vector<std::pair<Position, double> > getNeighbours(const Position pos,
			const std::vector<int> &map, int width, int height){
		std::vector<std::pair<Position, double> > ret;
		if(pos.x < width - 1 && map[Position(pos.x + 1, pos.y).toIndex(width)]){
			ret.push_back(std::make_pair(Position(pos.x + 1, pos.y), 1));
			if(pos.y < height - 1 && map[Position(pos.x, pos.y + 1).toIndex(width)]
					&& map[Position(pos.x + 1, pos.y + 1).toIndex(width)]){
				ret.push_back(std::make_pair(Position(pos.x + 1, pos.y + 1), DIAG_DIST));
			}
		}
		if(pos.x > 0 && map[Position(pos.x - 1, pos.y).toIndex(width)]){
			ret.push_back(std::make_pair(Position(pos.x - 1, pos.y), 1));
			if(pos.y > 0 && map[Position(pos.x, pos.y - 1).toIndex(width)]
					&& map[Position(pos.x - 1, pos.y - 1).toIndex(width)]){
				ret.push_back(std::make_pair(Position(pos.x - 1, pos.y - 1), DIAG_DIST));
			}
		}
		if(pos.y < height - 1 && map[Position(pos.x, pos.y + 1).toIndex(width)]){
			ret.push_back(std::make_pair(Position(pos.x, pos.y + 1), 1));
			if(pos.x > 0 && map[Position(pos.x - 1, pos.y).toIndex(width)]
					&& map[Position(pos.x - 1, pos.y + 1).toIndex(width)]){
				ret.push_back(std::make_pair(Position(pos.x - 1, pos.y + 1), DIAG_DIST));
			}
		}
		if(pos.y > 0 && map[Position(pos.x, pos.y - 1).toIndex(width)]){
			ret.push_back(std::make_pair(Position(pos.x, pos.y - 1), 1));
			if(pos.x < width - 1 && map[Position(pos.x + 1, pos.y).toIndex(width)]
					&& map[Position(pos.x + 1, pos.y - 1).toIndex(width)]){
				ret.push_back(std::make_pair(Position(pos.x + 1, pos.y - 1), DIAG_DIST));
			}
		}
		return ret;
	}

	virtual ~Pathfinder(){
	}

protected:

	enum class RootStrategy {
		GRID, FAR
	};

	std::vector<Position> roots;

	void spiral(Position &pos, std::vector<int> &map, int width, int height){
		int offset = 1, part = 0, startX = pos.x, startY = pos.y;
		while(!map[pos.toIndex(width)]){
			switch(part){
				case 0:
					if(pos.x + 1 == width){
						part++;
						break;
					}
					pos.x++;
					if(pos.x - startX == offset)
						part++;
					break;
				case 1:
					if(pos.y + 1 == height){
						part++;
						break;
					}
					pos.y++;
					if(pos.y - startY == offset)
						part++;
					break;
				case 2:
					if(pos.x == 0){
						part++;
						break;
					}
					pos.x--;
					if(pos.x - startX == -offset)
						part++;
					break;
				case 3:
					if(pos.y == 0){
						part = 0;
						offset++;
						break;
					}
					pos.y--;
					if(pos.y - startY == -offset){
						part = 0;
						offset++;
					}
					break;
			}
		}
	}

	void chooseRoots(std::vector<int> &map, int width, int height, int nRoots,
			RootStrategy strategy){
		roots.clear();
		int n, m;
		switch(strategy){
			case RootStrategy::GRID:
				if(nRoots != 1)//Add root of the single tree version for easy comparison
					chooseRoots(map, width, height, 1, RootStrategy::GRID);
				n = sqrt(nRoots);
				m = nRoots / n;
				while(n * m != nRoots){
					n--;
					m = nRoots / n;
				}
				n++;
				m++;
				if(width > height){
					int t = n;
					n = m;
					m = t;
				}
				for(int x = 1; x < n; x++){
					for(int y = 1; y < m; y++){
						Position pos((width * x) / n, (height * y) / m);
						spiral(pos, map, width, height);
						roots.push_back(pos);
					}
				}
				break;

			case RootStrategy::FAR:
				Position pos(0, 0);
				spiral(pos, map, width, height);
				roots.push_back(pos);
				for(int i = 1; i < nRoots; i++){
					double maxDist = -1;
					for(int x = 0; x < width; x++){
						for(int y = 0; y < height; y++){
							if(map[y * width + x]){
								double minDist = std::numeric_limits<double>::infinity();
								for(Position p : roots){
									double dist = sqrt(
											(p.x - x) * (p.x - x) + (p.y - y) * (p.y - y));
									if(dist < minDist){
										minDist = dist;
									}
								}
								if(minDist > maxDist){
									n = x;
									m = y;
									maxDist = minDist;
								}
							}
						}
					}
					roots.push_back(Position(n, m));
				}
				break;
		}
		std::sort(roots.begin(), roots.end());
		roots.erase(std::unique(roots.begin(), roots.end()), roots.end());
	}
};

#endif //PATHFINDER_H
